from .attached_files import Attachments
from .comments import Comments
from .issues import Issues
from .priorities import Priorities
from .worklogs import Worklogs

__all__ = [
    "Attachments",
    "Comments",
    "Issues",
    "Priorities",
    "Worklogs",
]



from __future__ import annotations

__all__ = ["Transitions"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

    from .transition import Transition


class Transitions(dict):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.current = -1

    def __iter__(self) -> Iterator[Transition]:
        """Iterate via transitions."""
        return self

    def __next__(self) -> Transition:
        """Get next transition."""
        self.current += 1
        values = list(self.values())
        if self.current >= len(values):
            raise StopIteration
        return values[self.current]



"""Base aiohttp client class module."""

from __future__ import annotations

import asyncio
import io
import logging
import ssl
from abc import ABC, abstractmethod
from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

import certifi
import msgspec
from aiohttp import BytesPayload, ClientSession, ClientTimeout, FormData, TCPConnector

from yatracker.exceptions import (
    AlreadyExistsError,
    NotAuthorizedError,
    ObjectNotFoundError,
    SufficientRightsError,
    YaTrackerError,
)

if TYPE_CHECKING:
    from aiohttp.typedefs import StrOrURL

DEFAULT_API_HOST = "https://api.tracker.yandex.net"
DEFAULT_API_VERSION = "v2"

logger = logging.getLogger(__name__)


class BaseClient(ABC):
    """Represents abstract base class for tracker client."""

    # ruff: noqa: PLR0913
    def __init__(
        self,
        org_id: str | int,
        token: str,
        headers: dict[str, str] | None = None,
        api_host: str | None = None,
        api_version: str | None = None,
        # ruff: noqa: ARG002
        **kwargs,
    ) -> None:
        """Set defaults on object init.

        By default, `self._session` is None.
        It will be created on a first API request.
        The second request will use the same `self._session`.
        """
        self._api_version = api_version or DEFAULT_API_VERSION
        self._base_url = api_host or DEFAULT_API_HOST
        _headers = headers.copy() if headers else {}
        _headers.setdefault("X-Org-Id", str(org_id))
        _headers.setdefault("Authorization", f"OAuth {token}")
        self._headers: dict[str, str] = _headers
        self._session: ClientSession | None = None
        self._encoder = msgspec.json.Encoder()

    async def request(
        self,
        method: str,
        uri: str,
        params: dict[str, Any] | None = None,
        payload: dict[str, Any] | None = None,
        form: FormData | None = None,
        **kwargs,
    ) -> bytes:
        """Make request."""
        if form:
            bytes_payload = form
        else:
            bytes_payload = BytesPayload(
                value=self._encoder.encode(payload),
                content_type="application/json",
            )

        # to support full links (e.g. Transition)
        if not uri.startswith("http"):
            uri = f"{self._base_url}/{self._api_version}{uri}"

        status, body = await self._make_request(
            method=method,
            url=uri,
            params=params,
            data=bytes_payload,
            **kwargs,
        )
        self._check_status(status, body)
        return body

    @abstractmethod
    async def _make_request(
        self,
        method: str,
        url: StrOrURL,
        **kwargs,
    ) -> tuple[int, bytes]:
        """Get raw response from via http-client.

        :returns: tuple of (status_code, response_body).
        """

    @staticmethod
    def _check_status(status: int, body: bytes) -> None:
        if status < HTTPStatus.MULTIPLE_CHOICES:
            return

        if status == HTTPStatus.UNAUTHORIZED:
            raise NotAuthorizedError

        if status == HTTPStatus.FORBIDDEN:
            raise SufficientRightsError

        if status == HTTPStatus.NOT_FOUND:
            raise ObjectNotFoundError

        if status == HTTPStatus.CONFLICT:
            raise AlreadyExistsError

        raise YaTrackerError(body)

    @abstractmethod
    async def close(self) -> None:
        """Close the session gracefully."""


class AIOHTTPClient(BaseClient):
    """Base aiohttp client.

    Consists of all methods need to make a request to API:
     - session caching
     - request wrapping
     - exceptions wrapping
     - grace session close
     - e.t.c.
    """

    # ruff: noqa: PLR0913
    def __init__(
        self,
        org_id: str | int,
        token: str,
        headers: dict[str, str] | None = None,
        api_host: str | None = None,
        api_version: str | None = None,
        **kwargs,
    ) -> None:
        """Set defaults on object init.

        By default, `self._session` is None.
        It will be created on a first API request.
        The second request will use the same `self._session`.
        """
        super().__init__(
            org_id=org_id,
            token=token,
            headers=headers,
            api_host=api_host,
            api_version=api_version,
            **kwargs,
        )
        self._timeout: ClientTimeout = kwargs.get("timeout") or ClientTimeout(total=0)

    def get_session(self) -> ClientSession:
        """Get cached session. One session per instance."""
        if isinstance(self._session, ClientSession) and not self._session.closed:
            return self._session

        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = TCPConnector(ssl=ssl_context)

        encoder = msgspec.json.Encoder()
        self._session = ClientSession(
            connector=connector,
            headers=self._headers,
            json_serialize=lambda obj: encoder.encode(obj).decode(),
            timeout=self._timeout,
        )
        return self._session

    async def _make_request(
        self,
        method: str,
        url: StrOrURL,
        **kwargs,
    ) -> tuple[int, bytes]:
        """Make a request.

        :param method: HTTP Method
        :param url: endpoint link
        :param kwargs: data, params, json and other...
        :return: status and result or exception
        """
        session = self.get_session()

        async with session.request(method, url, **kwargs) as response:
            status = response.status
            body = await response.read()

        if status >= HTTPStatus.BAD_REQUEST:
            raise self._process_exception(status, body)

        return status, body

    def _prepare_form(self, file: str | Path | io.IOBase) -> FormData:
        """Create form to pass file via multipart/form-data."""
        form = FormData()
        form.add_field("file", self._prepare_file(file))
        return form

    @staticmethod
    def _prepare_file(file: str | Path | io.IOBase) -> io.IOBase | BinaryIO:
        """Prepare accepted types to correct file type."""
        if isinstance(file, str):
            with Path(file).open("rb") as f:
                return f

        if isinstance(file, io.IOBase):
            return file

        if isinstance(file, Path):
            return file.open("rb")

        msg = (  # type: ignore[unreachable]
            f"Not supported file type: `{type(file).__name__}`"
        )
        raise TypeError(msg)

    @staticmethod
    def _process_exception(
        status: int,
        data: bytes,
    ) -> YaTrackerError:
        """Wrap API exceptions.

        :param status: response status
        :param data: response json converted to dict()
        :return: wrapped exception
        """
        text = data.decode("utf-8")
        logger.warning("Error! Status: %s. Body: %s", status, text)
        return YaTrackerError(text)

    async def close(self) -> None:
        """Close the session gracefully."""
        if not isinstance(self._session, ClientSession):
            return

        if self._session.closed:
            return

        await self._session.close()
        await asyncio.sleep(0.25)



from __future__ import annotations

__all__ = ["Transition"]

from .base import Base, field
from .status import Status


class Transition(Base, kw_only=True):
    id: str
    url: str = field(name="self")
    display: str
    to: Status

    async def execute(self) -> None:
        """Execute transition."""
        await self._tracker.execute_transition(self)



from __future__ import annotations

import logging
from functools import lru_cache
from typing import TYPE_CHECKING, Any, TypeVar

import msgspec.json

from yatracker.types.base import Base
from yatracker.utils.camel_case import camel_case

from .client import AIOHTTPClient

if TYPE_CHECKING:
    from collections.abc import Collection
    from types import TracebackType

    from .client import BaseClient

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseTracker:
    """Represents technical methods for using YaTracker."""

    # ruff: noqa: PLR0913
    def __init__(
        self,
        org_id: str | int | None = None,
        token: str | None = None,
        client: BaseClient | None = None,
        api_host: str | None = None,
        api_version: str | None = None,
    ) -> None:
        if (org_id is None or token is None) and client is None:
            msg = (
                "You must provide either `org_id` and `token` or `BaseClient` "
                "instance with set up headers `X-Org-Id` and `Authorization` and "
                "base url."
            )
            raise RuntimeError(msg)

        if client is not None:
            self._client = client
        else:
            if token is None or org_id is None:
                msg = "You must provide `org_id` and `token`."
                raise RuntimeError(msg)

            self._client = AIOHTTPClient(
                org_id=org_id,
                token=token,
                api_host=api_host,
                api_version=api_version,
            )

    def _decode(self, type_: type[T], data: bytes) -> T:
        """Decode bytes object to struct.

        Also add producer client object to `_tracker` field.
        """
        decoder = _get_decoder(type_)  # type: ignore[arg-type]
        obj = decoder.decode(data)
        _add_tracker(self, obj)
        return obj

    @staticmethod
    def clear_payload(
        payload: dict[str, Any],
        exclude: Collection[str] | None = None,
    ) -> dict[str, Any]:
        """Remove empty fields from payload."""
        payload = payload.copy()
        exclude = exclude or []
        kwargs = payload.pop("kwargs", None)
        if kwargs:
            payload.update(kwargs)

        return {
            camel_case(k): v
            for k, v in payload.items()
            if k not in {"self", "cls", *exclude} and v is not None
        }

    async def close(self) -> None:
        """Close gracefully."""
        await self._client.close()

    # ruff: noqa: PYI034
    async def __aenter__(self) -> BaseTracker:
        """Return async Tracker with async context."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        """Close async context."""
        await self.close()


@lru_cache
def _get_decoder(type_: type[T]) -> msgspec.json.Decoder:
    """Get cached msgspec encoder."""
    return msgspec.json.Decoder(type_)


def _add_tracker(tracker: BaseTracker, obj: Any) -> None:  # noqa: ANN401
    """Add tracker link to the object."""
    match obj:
        case Base():
            obj._tracker = tracker  # noqa: SLF001
        case list():
            for o in obj:
                _add_tracker(tracker, o)
        case dict():
            for v in obj.values():
                _add_tracker(tracker, v)



from __future__ import annotations

import logging

from .base import BaseTracker
from .categories import (
    Attachments,
    Comments,
    Issues,
    Priorities,
    Worklogs,
)

logger = logging.getLogger(__name__)


class YaTracker(
    Issues,
    Comments,
    Priorities,
    Attachments,
    Worklogs,
    BaseTracker,
):
    """Represents Yandex Tracker API client.

    API docs: https://cloud.yandex.com/en/docs/tracker/about-api

    Attention!
        All 'self' properties renamed to 'link' because it's incompatible with Python.
        All camelCase properties renamed to pythonic_case.
        Methods named by author, cause Yandex API has no clear method names.
        For help you to recognize method names full description is attached.

    """



from __future__ import annotations

__all__ = ["Worklog"]

from datetime import datetime

from .base import Base, field
from .issue import Issue
from .user import User


class Worklog(Base, kw_only=True):
    url: str = field(name="self")
    id: int
    version: int
    issue: Issue
    created_by: User
    updated_by: User | None = None
    created_at: datetime
    updated_at: datetime | None = None
    start: datetime
    duration: str



from __future__ import annotations

from yatracker.tracker.base import BaseTracker
from yatracker.types import Priority


class Priorities(BaseTracker):
    # ruff: noqa: FBT001 FBT002
    async def get_priorities(self, localized: bool = True) -> list[Priority]:
        """Get priorities.

        Use this request to get a list of priorities for an issue.
        """
        params = {"localized": str(localized).lower()} if localized else None
        data = await self._client.request(
            method="GET",
            uri="/priorities",
            params=params,
        )
        return self._decode(list[Priority], data)



import asyncio
from pathlib import Path

from yatracker import YaTracker

# CAUTION! Don't store credentials in your code!
ORG_ID = ...
TOKEN = ...
FILE_PATH = ...
FILE_NAME = ...


async def main() -> None:
    """Run basic example.

    This way you may create, get and edit an issue.
    """
    # define tracker (once)
    tracker = YaTracker(ORG_ID, TOKEN)

    # upload temp file
    with Path.open(FILE_PATH, "rb") as file:
        attachment = await tracker.upload_temp_file(file, FILE_NAME)

    # create an issue with attachment
    issue = await tracker.create_issue(
        summary="New Issue",
        queue="KEY",
        attachment_ids=[attachment.id],
    )

    # attach another one... or the same! :)
    with Path.open(FILE_PATH, "rb") as file:
        await tracker.attach_file(
            issue_id=issue.id,
            file=file,
            filename=FILE_NAME,
        )

    # list attachments
    attachments = await tracker.get_attachments(issue.id)

    # and delete them all
    for att in attachments:
        await tracker.delete_attachment(issue.id, att.id)

    # don't forget to close tracker on app shutdown (once)
    await tracker.close()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

__all__ = ["Priority"]

from .base import Base, field


class Priority(Base, kw_only=True):
    """Represents Priority.

    Attributes
    ----------
    url - Reference to the object.
    id - Priority ID.
    key - Priority key.
    version - Priority version.
    name - Display name of the priority. When localized=false is passed
    in the request, this parameter contains duplicates of
    the names in other languages.
    order - The weight of the priority. This parameter affects the order
    for displaying the priority in the interface.
    """

    url: str = field(name="self")
    id: str
    key: str
    display: str | None = None
    version: int | None = None
    name: str | dict | None = None
    order: int | None = None



from __future__ import annotations

from typing import Any

from yatracker.tracker.base import BaseTracker
from yatracker.types import (
    FullIssue,
    Issue,
    IssueType,
    Priority,
    Transition,
    Transitions,
)


class Issues(BaseTracker):
    async def get_issue(self, issue_id: str, expand: str | None = None) -> FullIssue:
        """Get issue parameters.

        Use this request to get information about an issue.

        :param issue_id: ID or key of the current issue.
        :param expand: Additional fields to include in the response:
                        transitions — Workflow transitions between statuses.
                        attachments — Attachments

        :return:
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}",
            params={"expand": expand} if expand else None,
        )
        return self._decode(FullIssue, data)

    async def edit_issue(
        self,
        issue_id: str,
        version: str | int | None = None,
        **kwargs,
    ) -> FullIssue:
        """Make changes to an issue.

        Use this request to make changes to an issue.
        The issue is selected by its ID or key.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/patch-issue
        """
        data = await self._client.request(
            method="PATCH",
            uri=f"/issues/{issue_id}",
            params={"version": str(version)} if version else None,
            payload=self.clear_payload(kwargs),
        )
        return self._decode(FullIssue, data)

    # ruff: noqa: ARG002 PLR0913
    async def create_issue(
        self,
        summary: str,
        queue: str,
        parent: Issue | None = None,
        description: str | None = None,
        sprint: dict[str, str] | None = None,
        type_: IssueType | None = None,
        priority: int | str | Priority | None = None,
        followers: list[str] | None = None,
        unique: str | None = None,
        attachment_ids: list[str] | None = None,
        **kwargs,
    ) -> FullIssue:
        """Create an issue."""
        payload = self.clear_payload(locals())
        data = await self._client.request(
            method="POST",
            uri="/issues/",
            payload=payload,
        )
        return self._decode(FullIssue, data)

    async def move_issue(
        self,
        issue_id: str,
        queue_key: str,
        *,
        notify: bool = True,
        notify_author: bool = False,
        move_all_fields: bool = False,
        initial_status: bool = False,
        expand: str | None = None,
        **kwargs,
    ) -> FullIssue:
        """Move an issue to a different queue.

        Before executing the request, make sure the user has permission
        to edit the issues to be moved and is allowed to create them in
        the new queue.

        Warning!
        If an issue you want to move has a type and status that are
        missing in the target queue, no transfer will be made. To reset
        the issue status to the initial value when moving it, use the
        InitialStatus parameter.

        By default, when an issue is moved, the values of its
        components, versions, and projects are cleared. If the new queue
        has the same values of the fields specified, use the
        MoveAllFields parameter to move the components, versions, and
        projects.

        If the issue has the local field values specified, they will be
        reset when moving the issue to a different queue.

        You can use the request body if you need to change the
        parameters of the issue being moved. The request body has the
        same format as when editing issues.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/move-issue
        """
        params: dict[str, Any] = {"queue": queue_key}

        if notify is False:
            params["notify"] = notify

        if notify_author is True:
            params["notifyAuthor"] = notify_author

        if move_all_fields is True:
            params["moveAllFields"] = move_all_fields

        if initial_status is True:
            params["initialStatus"] = initial_status

        if expand:
            params["expand"] = expand

        data = await self._client.request(
            method="POST",
            uri=f"/issues/{issue_id}/_move",
            params=params,
            payload=self.clear_payload(kwargs),
        )
        return self._decode(FullIssue, data)

    async def count_issues(
        self,
        filter_: dict[str, str] | None = None,
        query: str | None = None,
    ) -> int:
        """Get the number of issues.

        Use this request to find out how many issues meet the criteria in your request.
        :return:
        """
        payload: dict[str, Any] = {}
        if filter_ is not None:
            payload["filter"] = filter_
        if query is not None:
            payload["query"] = query

        data = await self._client.request(
            method="POST",
            uri="/issues/_count",
            payload=payload,
        )
        return self._decode(int, data)

    async def find_issues(
        self,
        filter_: dict[str, str] | None = None,
        query: str | None = None,
        order: str | None = None,
        expand: str | None = None,
        keys: str | None = None,
        queue: str | None = None,
    ) -> list[FullIssue]:
        """Find issues.

        Use this request to get a list of issues that meet specific criteria.
        If there are more than 10,000 issues in the response, use paging.
        :return:
        """
        payload = self.clear_payload(locals(), exclude=["expand", "order"])

        params = {}
        if order:
            params["order"] = order
        if expand:
            params["expand"] = expand

        data = await self._client.request(
            method="POST",
            uri="/issues/_search",
            params=params,
            payload=payload,
        )
        return self._decode(list[FullIssue], data)

    async def get_issue_links(self, issue_id: str) -> list[FullIssue]:
        """Get issue links.

        Use this request to get information about links between issues.
        The issue is selected by its ID or key.
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/links",
        )
        return self._decode(list[FullIssue], data)

    async def get_transitions(self, issue_id: str) -> Transitions:
        """Get transitions.

        Use this request to get a list of possible transitions for an issue.
        The issue is selected by its ID or key.
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/transitions",
        )
        transitions = self._decode(list[Transition], data)
        return Transitions(**{t.id: t for t in transitions})

    async def execute_transition(
        self,
        transition: Transition,
        **kwargs,
    ) -> list[Transition]:
        """Execute transition."""
        payload = self.clear_payload(kwargs)
        data = await self._client.request(
            method="POST",
            uri=f"{transition.url}/_execute",
            payload=payload,
        )
        return self._decode(list[Transition], data)



from __future__ import annotations

__all__ = ["FullIssue"]


from datetime import datetime

from .base import Base, field
from .comment import Comment
from .issue import Issue
from .issue_type import IssueType
from .priority import Priority
from .queue import Queue
from .sprint import Sprint
from .status import Status
from .transitions import Transitions
from .user import User


class FullIssue(Base, kw_only=True):
    url: str = field(name="self")
    id: str
    key: str
    version: int

    summary: str
    parent: Issue | None = None
    description: str | None = None
    sprint: list[Sprint] | None = None
    type: IssueType
    priority: Priority
    followers: list[User] | None = None
    queue: Queue
    previous_queue: Queue | None = None
    favorite: bool
    assignee: User | None = None

    last_comment_update_at: datetime | None = None
    aliases: list[str] | None = None
    updated_by: User | None = None
    created_at: datetime
    created_by: User
    votes: int
    updated_at: datetime | None = None
    status: Status
    previous_status: Status | None = None
    direction: str | None = None

    async def get_transitions(self) -> Transitions:
        """Return dict and list-like Transitions object.

        Iterate Transitions like a list:
        >>> transitions = await self.get_transitions()
        >>> for t in transitions:
        >>>    print(t)

        Use Transitions like a dict with transition names:
        >>> transitions = await self.get_transitions()
        >>> close = transitions.get('close')
        >>> if close:
        >>>    await close.execute()
        """
        return await self._tracker.get_transitions(self.id)

    async def get_comments(self) -> list[Comment]:
        """Get comments for self.

        :return:
        """
        return await self._tracker.get_comments(self.id)

    async def post_comment(self, text: str, **kwargs) -> Comment:
        """Post comment for self."""
        return await self._tracker.post_comment(self.id, text=text, **kwargs)



from __future__ import annotations

__all__ = ["IssueType"]

from .base import Base, field


class IssueType(Base, kw_only=True):
    url: str = field(name="self")
    id: str
    key: str
    display: str



from __future__ import annotations


class Printable:
    display: str | None

    def __str__(self) -> str:
        """Return display name."""
        try:
            return self.display or self.__class__.__name__
        except AttributeError:
            return super().__str__()



__all__ = [
    "Attachment",
    "Comment",
    "Duration",
    "FullIssue",
    "Issue",
    "IssueType",
    "Priority",
    "Queue",
    "Sprint",
    "Status",
    "Transition",
    "Transitions",
    "User",
    "Worklog",
]

from .attachment import Attachment
from .comment import Comment
from .duration import Duration
from .full_issue import FullIssue
from .issue import Issue
from .issue_type import IssueType
from .priority import Priority
from .queue import Queue
from .sprint import Sprint
from .status import Status
from .transition import Transition
from .transitions import Transitions
from .user import User
from .worklog import Worklog



import asyncio

from yatracker import YaTracker

# CAUTION! Don't store credentials in your code!
ORG_ID = ...
TOKEN = ...


async def main() -> None:
    """Run basic example.

    This way you may create, get and edit an issue.
    """
    # define tracker (once)
    tracker = YaTracker(ORG_ID, TOKEN)

    # create an issue
    issue = await tracker.create_issue("New Issue", "KEY")
    print(issue)

    # get the issue
    issue = await tracker.get_issue("KEY-1")
    print(issue)

    # update the issue
    issue = await tracker.edit_issue("KEY-1", description="Hello World")
    print(issue)

    # don't forget to close tracker on app shutdown (once)
    await tracker.close()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from datetime import datetime
from warnings import warn

from yatracker.tracker.base import BaseTracker
from yatracker.types import Worklog
from yatracker.types.duration import Duration


class Worklogs(BaseTracker):
    async def post_worklog(
        self,
        issue_id: str,
        start: str | datetime,
        duration: str | Duration,
        comment: str | None = None,
    ) -> Worklog:
        """Add worklog to the issue.

        Source:
        https://cloud.yandex.ru/docs/tracker/concepts/issues/new-worklog
        """
        if isinstance(start, datetime):
            start = start.isoformat(timespec="milliseconds")

        if isinstance(duration, Duration):
            duration = duration.to_iso()

        payload = self.clear_payload(locals(), exclude=["issue_id"])
        data = await self._client.request(
            method="POST",
            uri=f"/issues/{issue_id}/worklog/",
            payload=payload,
        )
        return self._decode(Worklog, data)

    async def edit_worklog(
        self,
        issue_id: str,
        worklog_id: int,
        duration: str | Duration,
        comment: str | None = None,
    ) -> Worklog:
        """Edit worklog.

        Source:
        https://cloud.yandex.ru/docs/tracker/concepts/issues/patch-worklog
        """
        if isinstance(duration, Duration):
            duration = duration.to_iso()

        query_params = ["issue_id", "worklog_id"]
        payload = self.clear_payload(locals(), exclude=query_params)
        data = await self._client.request(
            method="PATCH",
            uri=f"/issues/{issue_id}/worklog/{worklog_id}",
            payload=payload,
        )
        return self._decode(Worklog, data)

    async def delete_worklog(
        self,
        issue_id: str,
        worklog_id: int,
    ) -> bool:
        """Delete worklog.

        Source:
        https://cloud.yandex.ru/docs/tracker/concepts/issues/delete-worklog
        """
        await self._client.request(
            method="DELETE",
            uri=f"/issues/{issue_id}/worklog/{worklog_id}",
        )
        return True

    async def get_issue_worklog(self, issue_id: str) -> list[Worklog]:
        """Get issue worklog records.

        Source:
        https://cloud.yandex.ru/docs/tracker/concepts/issues/issue-worklog
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/worklog",
        )
        return self._decode(list[Worklog], data)

    async def get_worklog(
        self,
        created_by: str | None = None,
        created_at_from: datetime | str | None = None,
        created_at_to: datetime | str | None = None,
    ) -> list[Worklog]:
        """Get issue worklog records.

        Source:
        https://cloud.yandex.ru/docs/tracker/concepts/issues/get-worklog
        """
        created_at = _process_created_at(created_at_from, created_at_to)
        payload = self.clear_payload(
            locals(),
            exclude=["created_at_from", "created_at_to"],
        )
        data = await self._client.request(
            method="POST",
            uri="/worklog/_search",
            payload=payload,
        )
        return self._decode(list[Worklog], data)


def _process_created_at(
    created_at_from: datetime | str | None = None,
    created_at_to: datetime | str | None = None,
) -> dict[str, str] | None:
    date_range = [created_at_from, created_at_to]
    if any(date_range) and not all(date_range):
        msg = "Set full range or not set it at all."
        raise ValueError(msg)

    if not created_at_from or not created_at_to:
        return None

    if isinstance(created_at_from, datetime):
        if created_at_from.tzinfo is None:
            warn(
                "Tracker API may works wrong with naive datetime. "
                "Please, use Timezone-Aware objects.",
                UserWarning,
                stacklevel=2,
            )
        created_at_from = created_at_from.isoformat(timespec="milliseconds")

    if isinstance(created_at_to, datetime):
        created_at_to = created_at_to.isoformat(timespec="milliseconds")

    created_at: dict[str, str] = {
        "from": created_at_from,
        "to": created_at_to,
    }
    return created_at



from __future__ import annotations

__all__ = ["Duration"]

import re
from dataclasses import dataclass, fields

PATTERN = re.compile(
    r"^P(?=\d+[YMWD])"
    r"((?P<years>\d+)Y)?"
    r"((?P<months>\d+)M)?"
    r"((?P<weeks>\d+)W)?"
    r"((?P<days>\d+)D)?"
    r"(?P<time>T(?=\d+[HMS])"
    r"((?P<hours>\d+)H)?"
    r"((?P<minutes>\d+)M)?"
    r"((?P<seconds>\d+)S)?)?$",
)


@dataclass
class Duration:
    years: int = 0
    months: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0

    def to_iso(self) -> str:
        """Convert Duration object to ISO data."""
        time = ""
        if self.hours:
            time = f"{time}{self.hours}H"
        if self.minutes:
            time = f"{time}{self.minutes}M"
        if self.seconds:
            time = f"{time}{self.seconds}S"

        duration = "P"
        if self.years:
            duration = f"{duration}{self.years}Y"
        if self.months:
            duration = f"{duration}{self.months}M"
        if self.days:
            duration = f"{duration}{self.days}D"
        if time:
            duration = f"{duration}T{time}"

        return duration

    @classmethod
    def from_iso(cls, duration: str, pattern: re.Pattern = PATTERN) -> Duration:
        """Create Duration object from ISO data."""
        result = pattern.match(duration)
        if result is None:
            msg = "Duration is not matched to ISO duration pattern."
            raise ValueError(msg)

        data: dict[str, int] = {}
        for field in fields(Duration):
            if value := result.group(field.name):
                data[field.name] = int(value)

        return cls(**data)



from __future__ import annotations

__all__ = ["Base", "field"]

from typing import Any

from msgspec import Struct, field

from .mixins import Printable


class Base(Printable, Struct, omit_defaults=True, rename="camel"):
    """Base structure class."""

    _tracker: Any = None  # this field will be filled after decoding



from __future__ import annotations

from yatracker.tracker.base import BaseTracker
from yatracker.types import (
    Comment,
)


class Comments(BaseTracker):
    async def get_comments(self, issue_id: str) -> list[Comment]:
        """Get the comments for an issue.

        Use this request to get a list of comments in the issue.
        :param issue_id:
        :return:
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/comments",
        )
        return self._decode(list[Comment], data)

    async def post_comment(self, issue_id: str, text: str, **kwargs) -> Comment:
        """Comment the issue."""
        payload = self.clear_payload(locals(), exclude=["issue_id"])
        data = await self._client.request(
            method="POST",
            uri=f"/issues/{issue_id}/comments/",
            payload=payload,
        )
        return self._decode(Comment, data)



def camel_case(string: str) -> str:
    """Convert string into camel case."""
    if not string:
        return string

    string = string.replace("_", "-")
    lst = string.split("-")
    for i in range(len(lst)):
        if i == 0:
            continue
        lst[i] = lst[i].capitalize()

    return "".join(lst)



class YaTrackerError(Exception):
    """Represents base Yandex Tracker error."""


class NotAuthorizedError(YaTrackerError):
    def __init__(self) -> None:
        super().__init__(
            "The user is not authorized. Check whether all the steps "
            "described in API access were completed.",
        )


class SufficientRightsError(YaTrackerError):
    def __init__(self) -> None:
        super().__init__(
            "You do not have sufficient rights to perform this action. "
            "Double-check permissions in the Tracker interface. You "
            "need the same permissions to perform the action via the "
            "API as in the interface.",
        )


class ObjectNotFoundError(YaTrackerError):
    def __init__(self) -> None:
        super().__init__(
            "The requested object was not found. You might have entered"
            " an incorrect object ID or key value.",
        )


class AlreadyExistsError(YaTrackerError):
    def __init__(self) -> None:
        super().__init__(
            "An issue with the same value of the unique parameter already exists.",
        )



import asyncio
from datetime import datetime
from zoneinfo import ZoneInfo

from yatracker import YaTracker
from yatracker.types import Duration

# CAUTION! Don't store credentials in your code!
ORG_ID = ...
TOKEN = ...


async def main() -> None:
    """Run worklogs example."""
    # define tracker (once)
    tracker = YaTracker(ORG_ID, TOKEN)

    # create an issue
    issue = await tracker.create_issue(
        summary="New Issue",
        queue="KEY",
    )

    # add worklog to issue
    worklog = await tracker.post_worklog(
        issue_id=issue.id,
        start=datetime.now(ZoneInfo("Europe/Moscow")),
        duration=Duration(hours=5),
        comment="Worklog testing",
    )

    # edit worklog
    worklog = await tracker.edit_worklog(
        issue_id=issue.id,
        worklog_id=worklog.id,
        duration=Duration(minutes=5),
    )

    # delete worklog
    await tracker.delete_worklog(issue.id, worklog.id)

    # don't forget to close tracker on app shutdown (once)
    await tracker.close()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from typing import BinaryIO

from aiohttp import FormData

from yatracker.tracker.base import BaseTracker
from yatracker.types import Attachment


class Attachments(BaseTracker):
    async def get_attachments(self, issue_id: str) -> list[Attachment]:
        """Get a list of files attached to an issue.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/get-attachments-list
        """
        data = await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/attachments",
        )
        return self._decode(list[Attachment], data)

    async def download_attachment(
        self,
        issue_id: str,
        attachment_id: str | int,
        filename: str,
    ) -> bytes:
        """Download file attached to an issue.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/get-attachment
        """
        return await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/attachments/{attachment_id}/{filename}",
        )

    async def download_thumbnail(
        self,
        issue_id: str,
        attachment_id: str | int,
    ) -> bytes:
        """Get thumbnails of image files attached to issues.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/get-attachment-preview
        """
        return await self._client.request(
            method="GET",
            uri=f"/issues/{issue_id}/thumbnails/{attachment_id}",
        )

    async def attach_file(
        self,
        issue_id: str,
        file: BinaryIO,
        filename: str | None = None,
    ) -> Attachment:
        """Attach a file to an issue.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/post-attachment
        """
        form = FormData()
        form.add_field("file_data", file)
        data = await self._client.request(
            method="POST",
            uri=f"/issues/{issue_id}/attachments",
            params={"filename": filename} if filename else None,
            form=form,
        )
        return self._decode(Attachment, data)

    async def upload_temp_file(
        self,
        file: BinaryIO,
        filename: str | None = None,
    ) -> Attachment:
        """Upload temporary file.

        Use this request to upload a file to Tracker first, and then
        attach it when creating an issue or adding a comment.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/temp-attachment
        """
        form = FormData()
        form.add_field("file_data", file)
        data = await self._client.request(
            method="POST",
            uri="/attachments/",
            params={"filename": filename} if filename else None,
            form=form,
        )
        return self._decode(Attachment, data)

    async def delete_attachment(self, issue_id: str, attachment_id: str | int) -> bool:
        """Delete attached file.

        Source:
        https://cloud.yandex.com/en/docs/tracker/concepts/issues/delete-attachment
        """
        await self._client.request(
            method="DELETE",
            uri=f"/issues/{issue_id}/attachments/{attachment_id}/",
        )
        return True



from __future__ import annotations

__all__ = ["Attachment"]


from datetime import datetime

from .base import Base, field
from .user import User


class Attachment(Base, kw_only=True):
    """Represents attachment object."""

    url: str = field(name="self")
    id: str
    name: str
    content: str
    thumbnail: str | None = None
    created_by: User
    created_at: datetime
    mimetype: str
    size: int
    metadata: Metadata | None = None
    comment_id: str | None = None


class Metadata(Base, kw_only=True):
    """Represents attachment metadata."""

    size: str | None = None
